#include <new>
#include <stdio.h>
#include <stdarg.h>
#include <squirrel.h>
#include <sqthread.h>
#include <sqcont.h>
#include <AS3.h>

// sqwrapper.cpp より
extern void trace(const char *format, ...);
extern void sq_pushas3(HSQUIRRELVM v, AS3_Val val); ///< 格納用
extern SQRESULT sq_getas3(HSQUIRRELVM v, int idx, AS3_Val *result); ///< 取得用
extern AS3_Val createArgs(HSQUIRRELVM v, int idx);
extern AS3_Val createClass(const char *nsName, const char *className);
extern AS3_Val createInstance(const char *nsName, const char *className, AS3_Val args=NULL);
extern AS3_Val callFunction(const char *nsName, const char *funcName, AS3_Val args=NULL);

// ----------------------------------------------------------------

/**
 * ログ出力用 for squirrel
 */
static void PrintFunc(HSQUIRRELVM v, const SQChar* format, ...)
{
	va_list args;
	va_start(args, format);
#if 0
	char msg[1024];
	vsnprintf(msg, sizeof msg-1, format, args);
	AS3_Val m = AS3_String(msg);
	AS3_Trace(m);
	AS3_Release(m);
#else
	vprintf(format, args);
#endif
	va_end(args);
}

// ----------------------------------------------------------------------------
// 追加機能
// ----------------------------------------------------------------------------

// 引数エラー
SQRESULT ERROR_INVALIDPARAM(HSQUIRRELVM v) {
	return sq_throwerror(v, _SC("invalid param"));
}

/**
 * トレースログ表示
 * @param ... 表示内容(文字列に変換されて表示されます)
 */
static SQRESULT getStage(HSQUIRRELVM vm)
{
	// ステージオブジェクト
	AS3_Val stage = AS3_Stage();
	sq_pushas3(vm, stage);
	AS3_Release(stage);
	return 1;
}

/**
 * トレースログ表示
 * @param ... 表示内容(文字列に変換されて表示されます)
 */
static SQRESULT traceSQ(HSQUIRRELVM vm)
{
	SQInteger nargs = sq_gettop(vm);
	for (int i=1;i<=nargs;i++) {
		const SQChar *msg;
		sq_getstring(vm, i, &msg);
		AS3_Val m = AS3_String(msg ? msg : "null");
		AS3_Trace(m);
		AS3_Release(m);
	}
	return SQ_OK;
}

/**
 * Flash クラスオブジェクト生成 (squirrel用)
 * @param namaspace
 * @param classname
 * @param args 引数指定
 */
static SQRESULT createClassSQ(HSQUIRRELVM vm)
{
	SQInteger nargs = sq_gettop(vm);
	if (nargs >= 3) {
		const char *nsName = NULL;
		const char *className = NULL;
		sq_getstring(vm, 2, &nsName);
		sq_getstring(vm, 3, &className);
		if (nsName && className) {
			AS3_Val classObj = createClass(nsName, className);
			if (classObj) {
				sq_pushas3(vm, classObj);
				AS3_Release(classObj);
				return 1;
			} else {
				return sq_throwerror(vm, _SC("failed to create class"));
			}
		}
	}
	return ERROR_INVALIDPARAM(vm);
}

/**
 * Flash オブジェクト生成 (squirrel用)
 * @param namaspace
 * @param classname
 * @param args 引数指定
 */
static SQRESULT createInstanceSQ(HSQUIRRELVM vm)
{
	SQInteger nargs = sq_gettop(vm);
	if (nargs >= 3) {
		const char *nsName = NULL;
		const char *className = NULL;
		sq_getstring(vm, 2, &nsName);
		sq_getstring(vm, 3, &className);
		if (nsName && className) {
			AS3_Val args = createArgs(vm, 4);
			AS3_Val instance = createInstance(nsName, className, args);
			if (instance) {
				//trace("create instance done and push");
				sq_pushas3(vm, instance);
				AS3_Release(instance);
				return 1;
			} else {
				return sq_throwerror(vm, _SC("failed to create instance"));
			}
		}
	}
	return ERROR_INVALIDPARAM(vm);
}

/**
 * Flash グローバルファンクション呼び出し (squirrel用)
 * @param namaspace
 * @param funcname
 * @param args 引数指定
 */
static SQRESULT callFunctionSQ(HSQUIRRELVM vm)
{
	SQInteger nargs = sq_gettop(vm);
	if (nargs >= 3) {
		const char *nsName = NULL;
		const char *funcName = NULL;
		sq_getstring(vm, 2, &nsName);
		sq_getstring(vm, 3, &funcName);
		if (nsName && funcName) {
			AS3_Val args = createArgs(vm, 4);
			AS3_Val ret = callFunction(nsName, funcName, args);
			if (ret) {
				sq_pushas3(vm, ret);
				AS3_Release(ret);
				return 1;
			} else {
				return SQ_OK;
			}
		}
	}
	return ERROR_INVALIDPARAM(vm);
}

/**
 * Flashからの関数呼び返し用ラッパー
 */
class FlashCaller {

public:
	/**
	 * クラスの登録
	 * @param v squirrel VM
	 */
	static void registerClass()	{
		HSQUIRRELVM v = sqobject::getGlobalVM();
		sq_pushroottable(v); // root

		sq_pushstring(v, _SC("FlashCaller"), -1);
		sq_newclass(v, false);

#define REGISTER(name) \
	sq_pushstring(v, _SC(#name), -1);\
	sq_newclosure(v, name, 0);\
	sq_createslot(v, -3);

		REGISTER(constructor);
		REGISTER(getFlashFunction);
		
		sq_createslot(v, -3); // 生成したクラスを登録
		sq_pop(v,1); // root
	};

protected:

	// リリーサ
	static SQRESULT release(SQUserPointer up, SQInteger size) {
		if (up) {
			FlashCaller *self = (FlashCaller*)up;
			self->~FlashCaller();
			sq_free(self, sizeof *self);
		}
		return SQ_OK;
	}
	
	/**
	 * コンストラクタ
	 * ネイティブオブジェクトのコンストラクタに引数として HSQUIRRELVM を渡す。
	 */
	static SQRESULT constructor(HSQUIRRELVM v) {
		FlashCaller *self = (FlashCaller*)sq_malloc(sizeof *self);
		new (self) FlashCaller(v);
		if (self) {
			SQRESULT result;
			if (SQ_SUCCEEDED(result = sq_setinstanceup(v, 1, self))) {;
				sq_setreleasehook(v, 1, release);
			} else {
				delete self;
			}
			return result;
		} else {
			return SQ_ERROR;
		}
	}
	
	// Flashからの呼び出し用
	static AS3_Val caller(void *data, AS3_Val params) {
		FlashCaller *self = (FlashCaller*)data;
		return self->call(params);
	}

	// このオブジェクトに関連づいた Flashのファンクションを生成して返す
	static SQRESULT getFlashFunction(HSQUIRRELVM v) {
		SQUserPointer up;
		if (SQ_SUCCEEDED(sq_getinstanceup(v, 1, &up, NULL))) {
			AS3_Val ret = AS3_Function(up, caller);
			sq_pushas3(v, ret);
			AS3_Release(ret);
			return 1;
		}
		return SQ_ERROR;
	}
	
protected:

	/// コンストラクタ
	FlashCaller(HSQUIRRELVM v) {
		func.getStack(v, 2);
	}

	/// デストラクタ
	virtual ~FlashCaller() {
		func.clear();
	}

	// 呼び出し処理の下請け
	AS3_Val call(AS3_Val args) {
		AS3_Val ret = NULL;
		HSQUIRRELVM vm = sqobject::getGlobalVM();
		func.push(vm); // func
		
		// 引数処理
		int num = 1;
		sq_pushroottable(vm); // this
		if (args) {
			AS3_Val lenvalue = AS3_GetS(args, "length");
			if (lenvalue) {
				int len = AS3_IntValue(lenvalue);
				for (int i=0; i<len; i++) {
					AS3_Val idx = AS3_Int(i);
					AS3_Val value = AS3_Get(args, idx);
					sq_pushas3(vm, value);
					AS3_Release(value);
					AS3_Release(idx);
					num++;
				}
			}
			AS3_Release(lenvalue);
		}
		if (SQ_SUCCEEDED(sq_call(vm, num, SQTrue, SQTrue))) {
			sq_getas3(vm, -1, &ret);
			sq_pop(vm, 1); // result
		}
		sq_pop(vm, 1); // func
		return ret;
	}

private:
	/// 関数参照用
	sqobject::ObjectInfo func;
};

// ----------------------------------------------------------------------------
// tick値処理
// ----------------------------------------------------------------------------

// flash.utils.getTimer 保持用
static AS3_Val getTimer = NULL;

/// タイマー初期化
void initTimer()
{
	if (getTimer == NULL) {
		AS3_Val ns = AS3_String("flash.utils");
		getTimer = AS3_NSGetS(ns, "getTimer");
		AS3_Release(ns);
	}
}

/// タイマー終了
void doneTimer()
{
	AS3_Release(getTimer);
	getTimer = NULL;
}

/// タイマー値取得
int getCurrentTick()
{
	int tick;
	AS3_Val ret = AS3_Call(getTimer, NULL, NULL);
	tick = AS3_IntValue(ret);
	AS3_Release(ret);
	return tick;
}

// ----------------------------------------------------------------------------
// メイン
// ----------------------------------------------------------------------------

/**
 * メイン
 */
int
main(int argc, const char **argv)
{
	// 初期化処理
	HSQUIRRELVM vm = sqobject::init();
	sq_setprintfunc(vm, PrintFunc);

	sqobject::registerContinuous();
	sqobject::Object::registerClass();
	sqobject::Thread::registerClass();
	sqobject::Thread::registerGlobal();
	FlashCaller::registerClass();

	// グローバルメソッド・ファンクション登録
	
	sq_pushroottable(vm);

	// トレース用
	sq_pushstring(vm, _SC("trace"), -1);
	sq_newclosure(vm, traceSQ, 0);
	sq_createslot(vm, -3); 

	// ステージ取得用
	sq_pushstring(vm, _SC("getStage"), -1);
	sq_newclosure(vm, getStage, 0);
	sq_createslot(vm, -3); 

	// Flashクラス生成用
	sq_pushstring(vm, _SC("createClass"), -1);
	sq_newclosure(vm, createClassSQ, 0);
	sq_createslot(vm, -3); 
	// Flashオブジェクト生成用
	sq_pushstring(vm, _SC("createInstance"), -1);
	sq_newclosure(vm, createInstanceSQ, 0);
	sq_createslot(vm, -3); 
	// Flashグローバルファンクション呼び出し
	sq_pushstring(vm, _SC("callFunction"), -1);
	sq_newclosure(vm, callFunctionSQ, 0);
	sq_createslot(vm, -3); 

	sq_pop(vm, 1);

	// ---------------------------
	// ActionScript 側への登録処理
	
	// 初期スクリプト起動
	sqobject::Thread::init();
	sqobject::Thread::fork("init.nut", argc, argv);

	initTimer();

	// メインループ
	int prevTick = getCurrentTick();
	while (true) {
		int tick = getCurrentTick();
		int diff = tick - prevTick;

		if (sqobject::Thread::main(diff) <= 0) {
			break;
		}
		sqobject::mainContinuous();
		prevTick = tick;
		flyield();
	}

	doneTimer();
	
	sqobject::doneContinuous();
	sqobject::Thread::done();
	sqobject::done();
	
	return 0;
}
